"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SkeldjsPathfinder = void 0;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const util_1 = require("@skeldjs/util");
const core_1 = require("@skeldjs/core");
const events_1 = require("@skeldjs/events");
const Grid_1 = require("./util/Grid");
const Node_1 = require("./util/Node");
const engine_1 = require("./engine");
const events_2 = require("./events");
/**
 * Represents a pathfinding utility for the {@link SkeldjsClient SkeldJS Client}.
 *
 * See {@link SkeldjsPathfinderEvents} for events to listen to.
 */
class SkeldjsPathfinder extends events_1.EventEmitter {
    constructor(client, config = {}) {
        super();
        this.client = client;
        this.config = config;
        this._tick = 0;
        this._moved = false;
        this.isPaused = false;
        this.client.on("room.fixedupdate", this._ontick.bind(this));
        this.client.on("player.move", this._handleMove.bind(this));
        this.client.on("player.leave", this._handleLeave.bind(this));
        this.client.on("room.gameend", () => {
            this.grid = undefined;
        });
        this.client.on("player.syncsettings", () => {
            this.grid = undefined;
        });
    }
    get snode() {
        var _a;
        if (!this.position)
            return null;
        return (_a = this.grid) === null || _a === void 0 ? void 0 : _a.nearest(this.position.x, this.position.y);
    }
    get dnode() {
        var _a;
        if (!this.destination)
            return null;
        return (_a = this.grid) === null || _a === void 0 ? void 0 : _a.nearest(this.destination.x, this.destination.y);
    }
    get position() {
        var _a;
        return (_a = this.transform) === null || _a === void 0 ? void 0 : _a.position;
    }
    get transform() {
        var _a;
        return (_a = this.myPlayer) === null || _a === void 0 ? void 0 : _a.transform;
    }
    get myPlayer() {
        return this.client.myPlayer;
    }
    getMapId() {
        var _a;
        if (this.client.lobbyBehaviour) {
            return -1;
        }
        return (_a = this.client.settings) === null || _a === void 0 ? void 0 : _a.map;
    }
    getVentForMap(ventId) {
        if (typeof this.getMapId() === "undefined")
            return undefined;
        switch (this.getMapId()) {
            case core_1.GameMap.TheSkeld:
                return core_1.TheSkeldVents[ventId];
            case core_1.GameMap.MiraHQ:
                return core_1.MiraHQVents[ventId];
            case core_1.GameMap.Polus:
                return core_1.PolusVents[ventId];
            case core_1.GameMap.AprilFoolsTheSkeld:
                return core_1.TheSkeldVents[ventId];
            case core_1.GameMap.Airship:
                return core_1.AirshipVents[ventId];
            default:
                return undefined;
        }
    }
    async _ontick() {
        var _a, _b, _c;
        this._tick++;
        if (this._tick % SkeldjsPathfinder.MovementInterval !== 0)
            return;
        if (typeof this.getMapId() === "undefined")
            return;
        if (!this.grid) {
            const buff = fs_1.default.readFileSync(path_1.default.resolve(__dirname, "../../data/build", "" + this.getMapId()));
            this.grid = Grid_1.Grid.fromBuffer(buff);
        }
        if (!this.snode || !this.dnode)
            return;
        if (this._moved ||
            !this.path ||
            this._tick % (this.config.recalculateEvery || 1) === 0) {
            await this.recalculate();
            this._moved = false;
        }
        if (this.isPaused)
            return;
        const next = (_a = this.path) === null || _a === void 0 ? void 0 : _a.shift();
        if (next) {
            const pos = this.grid.actual(next.x, next.y);
            const dist = this.position ? util_1.Vector2.dist(this.position, pos) : 0;
            const ev = await this.emit(new events_2.EngineMoveEvent(pos));
            if (!ev.canceled) {
                (_b = this.transform) === null || _b === void 0 ? void 0 : _b.move(pos.x, pos.y, new util_1.Vector2(dist * this.client.settings.playerSpeed));
            }
            if (((_c = this.path) === null || _c === void 0 ? void 0 : _c.length) === 0) {
                this._stop(true);
            }
        }
        else {
            this.destination = undefined;
            this.path = undefined;
        }
    }
    async recalculate() {
        if (!this.grid || !this.snode || !this.dnode)
            return;
        this.grid.reset();
        this.path = engine_1.getShortestPath(this.grid, this.snode, this.dnode);
        this.path = this.path.filter((_, i, ar) => i === 0 || i === ar.length - 1 || i % 3 === 0);
        await this.emit(new events_2.EngineRecalculateEvent(this.path.map((node) => this.grid.actual(node.x, node.y))));
    }
    pause() {
        this.isPaused = true;
        this.emit(new events_2.PathfinderPauseEvent);
    }
    start() {
        if (!this.destination)
            return;
        this.isPaused = false;
        this.emit(new events_2.PathfinderStartEvent(this.destination));
    }
    _stop(reached) {
        this.destination = undefined;
        if (!reached)
            this._moved = true;
        this.emit(new events_2.PathfinderStopEvent(reached));
        if (reached) {
            this.emit(new events_2.PathfinderEndEvent);
        }
    }
    stop() {
        this._stop(false);
    }
    _go(dest) {
        this.destination = new util_1.Vector2(dest);
        this._moved = true;
        this.start();
    }
    go(pos) {
        var _a;
        const vec = pos;
        if (vec.x) {
            this._go(vec);
            return;
        }
        if (pos instanceof Node_1.Node) {
            if (!this.grid) {
                return;
            }
            this.grid.actual(pos.x, pos.y);
            return;
        }
        const resolved = this.client.resolvePlayer(pos);
        if (resolved && resolved.hasSpawned) {
            const position = (_a = resolved.transform) === null || _a === void 0 ? void 0 : _a.position;
            if (!position) {
                throw new Error("Could not follow player: they not fully spawned.");
            }
            return this.go(position);
        }
    }
    vent(ventid) {
        const coords = this.getVentForMap(ventid);
        if (!coords)
            return;
        this.go(new util_1.Vector2(coords.position));
    }
    _handleMove(ev) {
        if (ev.player === this.following) {
            this.destination = new util_1.Vector2(ev.position);
            this._moved = true;
        }
    }
    _handleLeave(ev) {
        if (ev.player === this.following) {
            this._stop(false);
            this.following = undefined;
        }
    }
    follow(player) {
        const resolved = this.client.resolvePlayer(player);
        if (resolved && resolved.hasSpawned) {
            this.following = resolved;
        }
    }
}
exports.SkeldjsPathfinder = SkeldjsPathfinder;
SkeldjsPathfinder.MovementInterval = 6;
//# sourceMappingURL=Pathfinder.js.map